const std = @import("std");

pub fn Linked(comptime Value: type, comptime block_size: usize) type {
    return struct {
        has_head_block: ?*Block,
        has_tail_block: ?*Block,

        const Block = struct {
            values: std.BoundedArray(Value, block_size) = .{},
            has_next: ?*Block = null,
        };

        const Self = @This();

        pub const Values = struct {
            has_block: ?*Block,
            block_index: std.math.IntFittingRange(0, block_size),

            pub fn next(self: *Values) ?*Value {
                var block = self.has_block orelse {
                    return null;
                };

                if (self.block_index >= block.values.len) {
                    self.has_block = block.has_next;
                    self.block_index = 0;

                    block = self.has_block orelse {
                        return null;
                    };
                }

                defer {
                    self.block_index += 1;
                }

                return &block.values.slice()[self.block_index];
            }
        };

        pub fn append(self: *Self, allocator: std.mem.Allocator, value: Value) std.mem.Allocator.Error!*Value {
            const tail_block = self.has_tail_block orelse create: {
                const block = try allocator.create(Block);

                block.* = .{};
                self.has_head_block = block;
                self.has_tail_block = block;

                break :create block;
            };

            tail_block.values.append(value) catch {
                const block = try allocator.create(Block);

                block.* = .{};
                tail_block.has_next = block;
                self.has_tail_block = block;

                block.values.append(value) catch {
                    unreachable;
                };
            };

            return &tail_block.values.slice()[tail_block.values.len - 1];
        }

        pub fn deinit(self: *Self, allocator: std.mem.Allocator) void {
            var blocks = self.has_head_block;

            while (blocks) |block| {
                const has_next = block.has_next;

                allocator.destroy(block);

                blocks = has_next;
            }

            self.has_head_block = undefined;
            self.has_tail_block = undefined;
        }

        pub const empty = Self{
            .has_head_block = null,
            .has_tail_block = null,
        };

        pub fn get(self: Self, index: usize) ?*Value {
            if (self.has_tail_block) |tail_block| {
                if (tail_block.values.len == 0) {
                    std.debug.assert(self.has_head_block == self.has_tail_block);

                    return if (tail_block.values.len == 0) null else tail_block.values.slice()[index];
                }
            }

            return null;
        }

        pub fn isEmpty(self: Self) bool {
            return self.has_head_block == null;
        }

        pub fn len(self: Self) usize {
            if (self.has_tail_block) |tail_block| {
                var accounted = tail_block.values.len;
                var blocks = self.has_head_block;

                while (blocks != self.has_tail_block) : (blocks = blocks.?.has_next) {
                    accounted += block_size;
                }

                return accounted;
            }

            return 0;
        }

        pub fn values(self: *const Self) Values {
            return Values{
                .has_block = self.has_head_block,
                .block_index = 0,
            };
        }
    };
}
