const coral = @import("./coral.zig");

const std = @import("std");

pub fn Hashed(comptime Key: type, comptime Value: type, comptime traits: Traits(Key)) type {
    const load_max = 0.75;
    const max_int = std.math.maxInt(usize);

    return struct {
        entry_map: []?Entry,
        len: usize,

        const Entry = struct {
            key: Key,
            value: Value,

            fn writeInto(self: Entry, table: *Self) ?*Entry {
                const hash_max = @min(max_int, table.entry_map.len);
                var hashed_key = traits.hash(self.key) % hash_max;
                var iterations = @as(usize, 0);

                while (true) : (iterations += 1) {
                    std.debug.assert(iterations < table.entry_map.len);

                    const table_entry = &(table.entry_map[hashed_key] orelse {
                        table.entry_map[hashed_key] = .{
                            .key = self.key,
                            .value = self.value,
                        };

                        table.len += 1;

                        return &table.entry_map[hashed_key].?;
                    });

                    if (traits.areEqual(table_entry.key, self.key)) {
                        return null;
                    }

                    hashed_key = (hashed_key +% 1) % hash_max;
                }
            }
        };

        pub const Keys = struct {
            context: *const Self,
            iterations: usize,

            pub fn next(self: *Keys) ?Key {
                while (self.iterations < self.context.entry_map.len) {
                    defer {
                        self.iterations += 1;
                    }

                    if (self.context.entry_map[self.iterations]) |entry| {
                        return entry.key;
                    }
                }

                return null;
            }
        };

        const Self = @This();

        pub const Values = struct {
            context: *const Self,
            iterations: usize,

            pub fn next(self: *Values) ?*Value {
                while (self.iterations < self.context.entry_map.len) {
                    defer {
                        self.iterations += 1;
                    }

                    if (self.context.entry_map[self.iterations]) |*entry| {
                        return &entry.value;
                    }
                }

                return null;
            }
        };

        pub fn clear(self: *Self) void {
            for (self.entry_map) |*entry| {
                entry.* = null;
            }

            self.len = 0;
        }

        pub fn deinit(self: *Self) void {
            if (self.entry_map.len != 0) {
                coral.heap.allocator.free(self.entry_map);
            }

            self.* = undefined;
        }

        pub const empty = Self{
            .entry_map = &.{},
            .len = 0,
        };

        pub fn insert(self: *Self, key: Key, value: Value) std.mem.Allocator.Error!bool {
            try self.rehash(load_max);

            std.debug.assert(self.entry_map.len > self.len);

            const entry = Entry{
                .key = key,
                .value = value,
            };

            return entry.writeInto(self) != null;
        }

        pub fn get(self: Self, key: Key) ?*Value {
            if (!self.isEmpty()) {
                const hash_max = @min(max_int, self.entry_map.len);
                var hashed_key = traits.hash(key) % hash_max;
                var iterations = @as(usize, 0);

                while (iterations < self.len) : (iterations += 1) {
                    const entry = &(self.entry_map[hashed_key] orelse {
                        return null;
                    });

                    if (traits.areEqual(entry.key, key)) {
                        return &entry.value;
                    }

                    hashed_key = (hashed_key +% 1) % hash_max;
                }
            }

            return null;
        }

        pub fn isEmpty(self: Self) bool {
            return self.len == 0;
        }

        pub fn keys(self: *const Self) Keys {
            return .{
                .context = self,
                .iterations = 0,
            };
        }

        pub fn loadFactor(self: Self) f32 {
            return if (self.entry_map.len == 0) 1 else @as(f32, @floatFromInt(self.len)) / @as(f32, @floatFromInt(self.entry_map.len));
        }

        pub fn rehash(self: *Self, max_load: f32) std.mem.Allocator.Error!void {
            if (self.loadFactor() <= max_load) {
                return;
            }

            var table = empty;

            errdefer {
                table.deinit();
            }

            table.entry_map = allocate: {
                const min_len = @max(1, self.len);
                const table_size = min_len * 2;
                const zeroed_entry_map = try coral.heap.allocator.alloc(?Entry, table_size);

                errdefer {
                    coral.heap.allocator.free(zeroed_entry_map);
                }

                for (zeroed_entry_map) |*entry| {
                    entry.* = null;
                }

                break :allocate zeroed_entry_map;
            };

            for (self.entry_map) |maybe_entry| {
                if (maybe_entry) |entry| {
                    std.debug.assert(entry.writeInto(&table) != null);
                }
            }

            self.deinit();

            self.* = table;
        }

        pub fn remove(self: *Self, key: Key) ?coral.KeyValuePair(Key, Value) {
            const hash_max = @min(max_int, self.entry_map.len);
            var hashed_key = traits.hash(key) % hash_max;

            while (true) {
                const entry = &(self.entry_map[hashed_key] orelse continue);

                if (traits.areEqual(entry.key, key)) {
                    const original_entry = entry.*;

                    self.entry_map[hashed_key] = null;

                    return .{
                        .key = original_entry.key,
                        .value = original_entry.value,
                    };
                }

                hashed_key = (hashed_key +% 1) % hash_max;
            }
        }

        pub fn replace(self: *Self, key: Key, value: Value) std.mem.Allocator.Error!?coral.KeyValuePair(Key, Value) {
            try self.rehash(load_max);

            std.debug.assert(self.entry_map.len > self.len);

            {
                const hash_max = @min(max_int, self.entry_map.len);
                var hashed_key = traits.hash(key) % hash_max;

                while (true) {
                    const entry = &(self.entry_map[hashed_key] orelse {
                        self.entry_map[hashed_key] = .{
                            .key = key,
                            .value = value,
                        };

                        self.len += 1;

                        return null;
                    });

                    if (traits.areEqual(key, entry.key)) {
                        const original_entry = entry.*;

                        entry.* = .{
                            .key = key,
                            .value = value,
                        };

                        return .{
                            .key = original_entry.key,
                            .value = original_entry.value,
                        };
                    }

                    hashed_key = (hashed_key +% 1) % hash_max;
                }
            }
        }

        pub fn values(self: *const Self) Values {
            return .{
                .context = self,
                .iterations = 0,
            };
        }
    };
}

pub fn Traits(comptime Key: type) type {
    return struct {
        areEqual: fn (Key, Key) bool,
        hash: fn (Key) usize,
    };
}

pub fn scalarTraits(comptime Scalar: type) Traits(Scalar) {
    const traits = switch (@typeInfo(Scalar)) {
        .@"enum" => |@"enum"| struct {
            fn areEqual(a: Scalar, b: Scalar) bool {
                return a == b;
            }

            fn hash(value: Scalar) usize {
                return switch (@sizeOf(@"enum".tag_type) > @sizeOf(usize)) {
                    true => @intFromEnum(value) % std.math.maxInt(usize),
                    false => @intFromEnum(value),
                };
            }
        },

        .pointer => struct {
            fn areEqual(a: Scalar, b: Scalar) bool {
                return a == b;
            }

            fn hash(value: Scalar) usize {
                return @intFromPtr(value);
            }
        },

        .int => |int| struct {
            fn areEqual(a: Scalar, b: Scalar) bool {
                return a == b;
            }

            fn hash(value: Scalar) Scalar {
                return switch (int.bits > @bitSizeOf(usize)) {
                    true => value % std.math.maxInt(usize),
                    false => value,
                };
            }
        },

        else => {
            @compileError(std.fmt.comptimePrint("parameter `Scalar` must be a scalar type, not {s}", .{
                @typeName(Scalar),
            }));
        },
    };

    return .{
        .areEqual = traits.areEqual,
        .hash = traits.hash,
    };
}

pub const string_traits = init: {
    const strings = struct {
        fn areEqual(a: []const u8, b: []const u8) bool {
            return std.mem.eql(u8, a, b);
        }

        fn hash(value: []const u8) usize {
            return coral.hashes.djb2(usize, value);
        }
    };

    break :init Traits([]const u8){
        .areEqual = strings.areEqual,
        .hash = strings.hash,
    };
};
