const coral = @import("../coral.zig");

const std = @import("std");

pub const AccessQualifier = enum(u32) {
    read_only = 0,
    write_only = 1,
    read_write = 2,
};

pub const AddressingModel = enum(u32) {
    logical = 0,
};

pub const BuildError = std.mem.Allocator.Error || error{OutOfIds};

pub const Capability = enum(u32) {
    matrix = 0,
    shader = 1,
};

pub const Dim = enum(u32) {
    @"1d" = 0,
    @"2d" = 1,
    @"3d" = 2,
    cube = 4,
};

pub const ExecutionModel = enum(u32) {
    fragment = 4,
};

pub const FloatType = struct {
    id: u32,
    bits: u32,
};

pub const FunctionControl = packed struct(u32) {
    @"inline": bool = false,
    dont_inline: bool = false,
    pure: bool = false,
    @"const": bool = false,
    reserved: u28 = 0,
};

pub const FunctionType = struct {
    id: u32,
    parameter_types: []const *const Type,
    return_type: *const Type,
};

pub const ImageFormat = enum(u32) {
    unknown = 0,
};

pub const ImageType = struct {
    id: u32,
    sampled_type: *const Type,
    dimensions: Dim,
    depth: u32,
    arrayed: u32,
    multi_sampled: u32,
    sampled: u32,
    format: ImageFormat,
    access: AccessQualifier,
};

pub const IntType = struct {
    id: u32,
    bits: u32,
    is_signed: bool,
};

pub const MemoryModel = enum(u32) {
    glsl450 = 1,
};

pub const Module = struct {
    ids_assigned: u32 = 0,
    capabilities: []const Capability,
    memory_model: struct { AddressingModel, MemoryModel },
    entry_points: PtrTree(*const coral.shaders.Function, EntryPoint) = .empty,
    types: PtrTree(*allowzero const anyopaque, Type) = .empty,
    inputs: PtrTree(*const coral.shaders.Input, Value) = .empty,
    functions: PtrTree(*const coral.shaders.Function, Function) = .empty,
    constants: PtrTree([*:0]const u8, Constant) = .empty,

    pub const Block = struct {
        id: u32,
        variables: PtrTree(*const coral.shaders.Local, Value) = .empty,
        instructions: coral.list.Linked(Instruction, 8) = .empty,

        pub fn shaderExpression(self: *Block, module: *Module, function: *Function, arena: *std.heap.ArenaAllocator, shader_expression: *const coral.shaders.Expression) BuildError!*const Value {
            const arena_allocator = arena.allocator();

            switch (shader_expression.*) {
                .group_expression => |expression| {
                    return self.shaderExpression(module, function, arena, expression);
                },

                .constant => |constant| {
                    return module.shaderConstant(arena, constant);
                },

                .get_input => |input| {
                    return module.shaderInput(arena, input);
                },

                .get_parameter => |parameter| {
                    return function.parameters.get(parameter) orelse {
                        @panic("parameter does not exist in function scope");
                    };
                },

                .invoke => |invoke| {
                    const arguments = try arena.allocator().alloc(*const Value, invoke.argument_count);

                    {
                        var shader_arguments = invoke.arguments;

                        for (arguments) |*argument| {
                            argument.* = try self.shaderExpression(module, function, arena, shader_arguments.?.expression);
                            shader_arguments = shader_arguments.?.has_next;
                        }
                    }

                    const instruction = try self.instructions.append(arena_allocator, .{
                        .call_function = .{
                            .arguments = arguments,
                            .function = try module.shaderFunction(arena, invoke.function),

                            .result = .{
                                .type = try module.shaderValueType(arena, invoke.function.has_return_type),
                                .id = try module.nextId(),
                            },
                        },
                    });

                    return &instruction.call_function.result;
                },

                .add => |binary_operation| {
                    const instruction = try self.instructions.append(arena_allocator, .{
                        .add = .{
                            .operands = .{
                                try self.shaderExpression(module, function, arena, binary_operation.lhs_expression),
                                try self.shaderExpression(module, function, arena, binary_operation.rhs_expression),
                            },

                            .result = .{
                                .type = try module.shaderValueType(arena, binary_operation.type),
                                .id = try module.nextId(),
                            },
                        },
                    });

                    return &instruction.add.result;
                },

                .subtract => |binary_operation| {
                    const instruction = try self.instructions.append(arena_allocator, .{
                        .subtract = .{
                            .operands = .{
                                try self.shaderExpression(module, function, arena, binary_operation.lhs_expression),
                                try self.shaderExpression(module, function, arena, binary_operation.rhs_expression),
                            },

                            .result = .{
                                .type = try module.shaderValueType(arena, binary_operation.type),
                                .id = try module.nextId(),
                            },
                        },
                    });

                    return &instruction.subtract.result;
                },

                .multiply => |binary_operation| {
                    const instruction = try self.instructions.append(arena_allocator, .{
                        .add = .{
                            .operands = .{
                                try self.shaderExpression(module, function, arena, binary_operation.lhs_expression),
                                try self.shaderExpression(module, function, arena, binary_operation.rhs_expression),
                            },

                            .result = .{
                                .type = try module.shaderValueType(arena, binary_operation.type),
                                .id = try module.nextId(),
                            },
                        },
                    });

                    return &instruction.multiply.result;
                },

                .divide => |binary_operation| {
                    const instruction = try self.instructions.append(arena_allocator, .{
                        .add = .{
                            .operands = .{
                                try self.shaderExpression(module, function, arena, binary_operation.lhs_expression),
                                try self.shaderExpression(module, function, arena, binary_operation.rhs_expression),
                            },

                            .result = .{
                                .type = try module.shaderValueType(arena, binary_operation.type),
                                .id = try module.nextId(),
                            },
                        },
                    });

                    return &instruction.divide.result;
                },

                else => unreachable,
            }
        }

        pub fn shaderVariable(self: *Block, module: *Module, function: *Function, arena: *std.heap.ArenaAllocator, local: *const coral.shaders.Local) BuildError!*const Value {
            return self.variables.get(local) orelse {
                const arena_allocator = arena.allocator();

                const variable = (try self.variables.insert(arena_allocator, local, .{
                    .type = try module.shaderValueType(arena, local.type),
                    .id = try module.nextId(),
                })).?;

                _ = try self.instructions.append(arena_allocator, .{
                    .store = .{
                        .target = variable,
                        .source = try self.shaderExpression(module, function, arena, local.expression),
                    },
                });

                return variable;
            };
        }
    };

    pub const Constant = struct {
        value: Value,
        data: []const u32,
    };

    pub const EntryPoint = struct {
        id: u32,
        function: *const Function,
        execution_model: ExecutionModel,
        interfaces: PtrTree(*const Value, void) = .empty,
    };

    pub const Function = struct {
        id: u32,
        type: *const Type,
        control: FunctionControl,
        parameters: PtrTree(*const coral.shaders.Parameter, Value) = .empty,
        block: Block,
    };

    pub const Instruction = union(enum) {
        store: struct { target: *const Value, source: *const Value },
        call_function: FunctionCall,
        add: BinaryOp,
        subtract: BinaryOp,
        multiply: BinaryOp,
        divide: BinaryOp,

        pub const BinaryOp = struct {
            result: Value,
            operands: [2]*const Value,
        };

        pub const FunctionCall = struct {
            result: Value,
            function: *const Function,
            arguments: []*const Value,
        };
    };

    fn PtrTree(comptime Ptr: type, comptime Node: type) type {
        return coral.tree.Binary(Ptr, Node, coral.tree.scalarTraits(Ptr));
    }

    fn nextId(self: *Module) BuildError!u32 {
        self.ids_assigned = coral.scalars.add(self.ids_assigned, 1) orelse {
            return error.OutOfIds;
        };

        return self.ids_assigned;
    }

    pub fn shaderConstant(self: *Module, arena: *std.heap.ArenaAllocator, shader_constant: [:0]const u8) BuildError!*const Value {
        if (self.constants.get(shader_constant)) |constant| {
            return &constant.value;
        }

        const arena_allocator = arena.allocator();
        const data = try arena_allocator.alloc(u32, 1);

        if (std.mem.indexOfScalar(u8, shader_constant, '.') == null) {
            data[0] = std.mem.nativeToLittle(u32, @bitCast(coral.utf8.DecFormat.c.parse(i32, shader_constant) orelse {
                @panic("malformed constant");
            }));

            const constant = (try self.constants.insert(arena_allocator, shader_constant, .{
                .data = data,

                .value = .{
                    .type = try self.shaderValueType(arena, .int),
                    .id = try self.nextId(),
                },
            })).?;

            return &constant.value;
        } else {
            data[0] = std.mem.nativeToLittle(u32, @bitCast(coral.utf8.DecFormat.c.parse(f32, shader_constant) orelse {
                @panic("malformed constant");
            }));

            const constant = (try self.constants.insert(arena_allocator, shader_constant, .{
                .data = data,

                .value = .{
                    .type = try self.shaderValueType(arena, .float),
                    .id = try self.nextId(),
                },
            })).?;

            return &constant.value;
        }
    }

    pub fn shaderEntryPoint(
        self: *Module,
        arena: *std.heap.ArenaAllocator,
        execution_model: ExecutionModel,
        shader_function: *const coral.shaders.Function,
    ) BuildError!*const EntryPoint {
        return self.entry_points.get(shader_function) orelse {
            const arena_allocator = arena.allocator();

            const entry_point = (try self.entry_points.insert(arena_allocator, shader_function, .{
                .execution_model = execution_model,
                .function = try self.shaderFunction(arena, shader_function),
                .id = try self.nextId(),
            })).?;

            // try entry_point.registerBlockInterfaces(shader_function.block);

            return entry_point;
        };
    }

    pub fn shaderFunction(
        self: *Module,
        arena: *std.heap.ArenaAllocator,
        shader_function: *const coral.shaders.Function,
    ) BuildError!*const Function {
        return self.functions.get(shader_function) orelse {
            const arena_allocator = arena.allocator();

            const function = (try self.functions.insert(arena_allocator, shader_function, .{
                .control = .{},
                .type = try self.shaderFunctionType(arena, shader_function),
                .id = try self.nextId(),
                .block = .{ .id = try self.nextId() },
            })).?;

            {
                var parameters = shader_function.parameters;

                while (parameters) |parameter| : (parameters = parameter.has_next) {
                    _ = try function.parameters.insert(arena_allocator, parameter, .{
                        .type = try self.shaderValueType(arena, parameter.type),
                        .id = try self.nextId(),
                    });
                }
            }

            {
                var statements = shader_function.block.statements;

                while (statements) |statement| {
                    switch (statement.*) {
                        .declare_local => |local_declaration| {
                            _ = try function.block.shaderVariable(self, function, arena, local_declaration.local);
                            statements = local_declaration.has_next;
                        },

                        .mutate_local => {},
                        .mutate_output => {},

                        .return_expression => |_| {},
                    }
                }
            }

            return function;
        };
    }

    pub fn shaderFunctionType(
        self: *Module,
        arena: *std.heap.ArenaAllocator,
        shader_function: *const coral.shaders.Function,
    ) BuildError!*const Type {
        const function_signature_address: *allowzero const anyopaque = @ptrCast(shader_function.signature);

        return self.types.get(function_signature_address) orelse {
            const arena_allocator = arena.allocator();
            const parameter_types = try arena_allocator.alloc(*const Type, shader_function.parameter_count);

            {
                var parameters = shader_function.parameters;

                for (parameter_types) |*parameter_type| {
                    parameter_type.* = try self.shaderValueType(arena, parameters.?.type);
                    parameters = parameters.?.has_next;
                }
            }

            return (try self.types.insert(arena_allocator, function_signature_address, .{
                .function = .{
                    .parameter_types = parameter_types,
                    .return_type = try self.shaderValueType(arena, shader_function.has_return_type),
                    .id = try self.nextId(),
                },
            })).?;
        };
    }

    pub fn shaderInput(
        self: *Module,
        arena: *std.heap.ArenaAllocator,
        shader_input: *const coral.shaders.Input,
    ) BuildError!*const Value {
        return self.inputs.get(shader_input) orelse (try self.inputs.insert(arena.allocator(), shader_input, .{
            .type = try self.shaderValueType(arena, shader_input.type),
            .id = try self.nextId(),
        })).?;
    }

    pub fn shaderValueType(
        self: *Module,
        arena: *std.heap.ArenaAllocator,
        has_shader_type: ?*const coral.shaders.Type,
    ) BuildError!*const Type {
        const arena_allocator = arena.allocator();
        const symbol_address: *allowzero const anyopaque = @ptrCast(has_shader_type);

        return self.types.get(symbol_address) orelse {
            const id = try self.nextId();

            const symbol = has_shader_type orelse {
                return (try self.types.insert(arena_allocator, symbol_address, .{
                    .void = .{
                        .id = id,
                    },
                })).?;
            };

            switch (symbol.layout) {
                .signed, .unsigned => |scalar| {
                    return (try self.types.insert(arena_allocator, symbol_address, .{
                        .int = .{
                            .bits = scalar.bits,
                            .is_signed = (symbol.layout == .signed),
                            .id = id,
                        },
                    })).?;
                },

                .float => |float| {
                    return (try self.types.insert(arena_allocator, symbol_address, .{
                        .float = .{
                            .bits = float.bits,
                            .id = id,
                        },
                    })).?;
                },

                .vector => |vector| {
                    return (try self.types.insert(arena_allocator, symbol_address, .{
                        .vector = .{
                            .id = id,
                            .component_type = try self.shaderValueType(arena, vector.element),
                            .component_len = vector.dimensions.count(),
                        },
                    })).?;
                },

                .matrix, .record => unreachable,

                .texture => {
                    const image_type = try arena_allocator.create(Type);

                    image_type.* = .{
                        .image = .{
                            .depth = 0,
                            .arrayed = 0,
                            .multi_sampled = 0,
                            .sampled = 1,
                            .dimensions = .@"2d",
                            .format = .unknown,
                            .access = .read_only,
                            .sampled_type = try self.shaderValueType(arena, .float),
                            .id = try self.nextId(),
                        },
                    };

                    return (try self.types.insert(arena_allocator, symbol_address, .{
                        .sampled_image = .{
                            .image_type = image_type,
                            .id = try self.nextId(),
                        },
                    })).?;
                },
            }
        };
    }

    pub fn writeTo(self: Module, spirv: coral.bytes.Writable) coral.bytes.ReadWriteError!void {
        const Header = extern struct {
            magic_number: u32 = 0x07230203, // SPIR-V magic number in little-endian
            version: u32, // SPIR-V version (e.g., 0x00010000 for 1.0)
            generator_magic: u32, // Tool or generator identifier
            id_upper_bound: u32, // Upper bound for IDs
            reserved: u32 = 0, // Reserved, always 0
        };

        const OpInstruction = enum(u32) {
            memory_model = 14,
            entry_point = 15,
            execution_mode = 16,
        };

        try coral.bytes.writeLittle(spirv, Header{
            .magic_number = 0x07230203,
            .version = 0x00010000,
            .generator_magic = 0,
            .id_upper_bound = self.ids_assigned,
        });

        {
            const addressing, const memory = self.memory_model;

            try coral.bytes.writeLittle(spirv, OpInstruction.memory_model);
            try coral.bytes.writeLittle(spirv, addressing);
            try coral.bytes.writeLittle(spirv, memory);
        }

        {
            var entry_points = self.entry_points.keyValues();

            while (entry_points.next()) |entry_point| {
                try coral.bytes.writeLittle(spirv, OpInstruction.entry_point);
                try coral.bytes.writeLittle(spirv, entry_point.value.execution_model);
                try coral.bytes.writeLittle(spirv, entry_point.value.function.id);
                try writeString(spirv, entry_point.key.identifier);

                var interfaces = entry_point.value.interfaces.keyValues();

                while (interfaces.nextKey()) |interface| {
                    try coral.bytes.writeLittle(spirv, interface.*.id);
                }

                try coral.bytes.writeLittle(spirv, OpInstruction.execution_mode);
                try coral.bytes.writeLittle(spirv, entry_point.value.function.id);
                try coral.bytes.writeLittle(spirv, @as(u32, 7));
            }
        }
    }
};

pub const Value = struct {
    id: u32,
    type: *const Type,
};

pub const SampledImageType = struct {
    id: u32,
    image_type: *const Type,
};

pub const StructType = struct {
    id: u32,
    member_types: []const *const Type,
};

pub const Type = union(enum) {
    void: VoidType,
    float: FloatType,
    vector: VectorType,
    image: ImageType,
    @"struct": StructType,
    function: FunctionType,
    int: IntType,
    sampled_image: SampledImageType,
};

pub const VectorType = struct {
    id: u32,
    component_type: *const Type,
    component_len: u32,
};

pub const VoidType = struct {
    id: u32,
};

fn writeString(spirv: coral.bytes.Writable, data: []const u8) coral.bytes.ReadWriteError!void {
    const data_len_with_sentinel = data.len + 1;
    const word_size = @sizeOf(u32);
    const word_count = (data_len_with_sentinel + word_size) / word_size;

    try coral.bytes.writeSentineled(spirv, data, 0);

    const padding = (word_count * word_size) - data_len_with_sentinel;

    try coral.bytes.writeN(spirv, "\x00", padding);
}
