pub const Behavior = @import("./App/Behavior.zig");

pub const Setup = @import("./App/Setup.zig");

const SystemGraph = @import("./App/SystemGraph.zig");

const coral = @import("coral");

const ona = @import("./ona.zig");

const std = @import("std");

initialized_states: coral.map.Hashed(*const coral.TypeId, coral.Box, coral.map.scalarTraits(*const coral.TypeId)),
named_systems: coral.tree.Binary([]const u8, SystemGraph, coral.tree.sliceTraits([]const u8)),
is_running: bool,

pub const RunError = std.mem.Allocator.Error || error{
    MissingDependency,
};

const Self = @This();

pub const Time = struct {
    elapsed: f64,
};

pub fn deinit(self: *Self) void {
    {
        var key_values = self.named_systems.keyValues();

        while (key_values.nextValue()) |system| {
            system.deinit();
        }

        self.named_systems.deinit(coral.heap.allocator);
    }

    {
        var states = self.initialized_states.values();

        while (states.next()) |state| {
            state.deinit();
        }

        self.initialized_states.deinit();
    }
}

fn scheduleName(comptime schedule: anytype) [:0]const u8 {
    return switch (@typeInfo(@TypeOf(schedule))) {
        .enum_literal => @tagName(schedule),
        else => @compileError("paramater `schedule` must be an enum literal type"),
    };
}

pub fn hasState(self: *Self, comptime State: type) ?*State {
    if (self.initialized_states.get(.of(State))) |boxed_state| {
        return boxed_state.has(State).?;
    }

    return null;
}

pub fn init() std.mem.Allocator.Error!Self {
    var self = Self{
        .initialized_states = .empty,
        .named_systems = .empty,
        .is_running = true,
    };

    try self.setState(Time{
        .elapsed = 0,
    });

    return self;
}

pub fn on(self: *Self, comptime schedule: anytype, behavior: *const Behavior) std.mem.Allocator.Error!void {
    const schedule_name = scheduleName(schedule);
    const systems = self.named_systems.get(schedule_name) orelse (try self.named_systems.insert(coral.heap.allocator, schedule_name, .empty)).?;

    try systems.insert(self, behavior);
}

pub fn putState(self: *Self, state: anytype) std.mem.Allocator.Error!*@TypeOf(state) {
    try self.setState(state);

    return self.hasState(@TypeOf(state)).?;
}

pub fn run(self: *Self, tasks: *coral.asio.TaskQueue, comptime schedule: anytype) RunError!void {
    const schedule_name = scheduleName(schedule);
    const systems = self.named_systems.get(schedule_name) orelse (try self.named_systems.insert(coral.heap.allocator, schedule_name, .empty)).?;

    try systems.run(self, tasks);
}

pub fn setState(self: *Self, state: anytype) std.mem.Allocator.Error!void {
    var boxed_state = try coral.Box.init(state);

    errdefer {
        boxed_state.deinit();
    }

    var replaced = try self.initialized_states.replace(.of(@TypeOf(state)), boxed_state);

    if (replaced) |*entry| {
        entry.value.deinit();
    }
}
